/*
 * PeakThread.cpp
 *
 *  Created on: 08.06.2019
 *      Author: michi
 */

#include "PeakThread.h"
#include "../../data/Song.h"
#include "../../data/Track.h"
#include "../../data/TrackLayer.h"
#include "../../data/Sample.h"
#include "../../data/audio/AudioBuffer.h"
#include "../../stuff/PerformanceMonitor.h"
#include "../../lib/hui/hui.h"
#include "../../lib/os/time.h"


InterThreadMessager::~InterThreadMessager() {
	flush();
}

void InterThreadMessager::unsubscribe(VirtualBase *o) {
	flush();
	Observable::unsubscribe(o);
}

void InterThreadMessager::flush() {
	if (flushing.load())
		return;
	flushing = true;
	while (counter.load() > 0)
		hui::Application::do_single_main_loop();
		//os::sleep(0.01f);
}

// call from any thread, will notify in main/GUI thread
void InterThreadMessager::notify_x() {
	if (flushing.load())
		return;
	counter.fetch_add(1);
	hui::run_later(0.001f, [this] {
		if (!flushing.load())
			this->notify();
		counter.fetch_sub(1);
	});
}




PeakThread::PeakThread(Song *s) {
	song = s;
	allow_running = true;
	updating = false;
	perf_channel = PerformanceMonitor::create_channel("peakthread", this);
}

PeakThread::~PeakThread() {
	PerformanceMonitor::delete_channel(perf_channel);
}

void PeakThread::on_run() {
	while (allow_running) {
		if (updating) {
			try {
				update_song();
				updating = false;
				notify();
				//msg_write(":D");
			} catch(Exception &e) {
				//msg_write(":(    " + e.message());
			}
		}
		os::sleep(0.05f);
		Thread::cancelation_point();
	}
}

void PeakThread::start_update() {
	if (updating)
		stop_update();
	//msg_write("PT START");
	updating = true;
}

void PeakThread::stop_update() {
	//msg_write("PT STOP");
	updating = false;
}

void PeakThread::hard_stop() {
	//msg_write("PT HARD STOP");
	updating = false;
	allow_running = false;
	kill();
}

void PeakThread::update_buffer(AudioBuffer &buf) {
	song->lock();
	if (!updating) {
		song->unlock();
		throw Exception("aaa4");
	}
	int n = buf._update_peaks_prepare();
	song->unlock();

	Thread::cancelation_point();
	if (!updating)
		throw Exception("aaa3");

	for (int i=0; i<n; i++) {
		if (buf._peaks_chunk_needs_update(i)) {
			while (!song->try_lock()) {
				Thread::cancelation_point();
				os::sleep(0.01f);
				if (!updating)
					throw Exception("aaa");
			}
			PerformanceMonitor::start_busy(perf_channel);
			buf._update_peaks_chunk(i);
			PerformanceMonitor::end_busy(perf_channel);
			song->unlock();
			notify();
			Thread::cancelation_point();
		}
		if (!updating)
			throw Exception("aaa2");
	}
}

void PeakThread::update_track(Track *t) {
	for (TrackLayer *l: weak(t->layers))
		for (AudioBuffer &b: l->buffers)
			update_buffer(b);
}

void PeakThread::update_song() {
	//msg_write(".");
	for (Track *t: weak(song->tracks))
		update_track(t);
	for (Sample *s: weak(song->samples))
		if (s->buf)
			update_buffer(*s->buf);
}

void PeakThread::notify() {
	messanger.notify_x();
	//hui::run_later(0.01f, [this] { view->force_redraw(); });
}

